"""Common definitions for the tinydns test suite."""

from __future__ import annotations

import dataclasses
import subprocess  # noqa: S404
import typing


if typing.TYPE_CHECKING:
    import pathlib


@dataclasses.dataclass(frozen=True)
class TestRecord:
    """A single test record."""

    rtype: str
    name: str
    address: str
    answers: list[str]


@dataclasses.dataclass(frozen=True)
class RecordType:
    """Metadata about a record type."""

    query: str
    rtype: int
    prefix: str


MINENC = "us-ascii"
SAFEENC = "ISO-8859-1"

SVCDIR = "svc-tinydns"

RECORDS = [
    TestRecord(
        rtype="ns",
        name="example.com",
        address="127.6.16.1",
        answers=[
            "answer: example.com 259200 NS a.ns.example.com",
            "answer: example.com 259200 NS b.ns.example.com",
        ],
    ),
    TestRecord(
        rtype="ns",
        name="example.com",
        address="127.6.16.2",
        answers=[
            "answer: example.com 259200 NS a.ns.example.com",
            "answer: example.com 259200 NS b.ns.example.com",
        ],
    ),
    TestRecord(
        rtype="mx",
        name="example.com",
        address="127.6.16.1",
        answers=["answer: example.com 86400 MX 0 a.mx.example.com"],
    ),
    TestRecord(
        rtype="host",
        name="example.com",
        address="127.6.16.3",
        answers=["answer: example.com 86400 A 127.6.16.3"],
    ),
    TestRecord(
        rtype="alias",
        name="www.example.com",
        address="127.6.16.3",
        answers=["answer: www.example.com 86400 A 127.6.16.3"],
    ),
    TestRecord(
        rtype="childns",
        name="dev.example.com",
        address="127.6.16.17",
        answers=["authority: dev.example.com 259200 NS a.ns.dev.example.com"],
    ),
]

TYPES = {
    "ns": RecordType(query="ns", rtype=2, prefix="."),
    "mx": RecordType(query="mx", rtype=15, prefix="@"),
    "host": RecordType(query="a", rtype=1, prefix="="),
    "alias": RecordType(query="a", rtype=1, prefix="+"),
    "childns": RecordType(query="ns", rtype=2, prefix="&"),
}


@dataclasses.dataclass(frozen=True)
class Config:
    """Runtime configuration for the tinydns test program."""

    bindir: pathlib.Path
    sbindir: pathlib.Path
    subenv: dict[str, str]
    verbose: bool

    def diag(self, msg: str) -> None:
        """Display a diagnostic message if requested."""
        if self.verbose:
            print(f"diag: {msg}")

    def check_call(self, cmd: list[pathlib.Path | str]) -> None:
        """Run a program in a sane environment."""
        subprocess.check_call(cmd, shell=False, env=self.subenv)  # noqa: S603

    def check_output(self, cmd: list[pathlib.Path | str], encoding: str = MINENC) -> str:
        """Return the output of a program run in a sane environment."""
        return subprocess.check_output(  # noqa: S603
            cmd,
            shell=False,
            env=self.subenv,
            encoding=encoding,
        )



"""Test the `tinydns-data` tool and the produced CDB file."""

from __future__ import annotations

import re
import sys
import typing

from . import defs


if typing.TYPE_CHECKING:
    import pathlib


NULL = bytes([0]).decode(defs.SAFEENC)

RE_CDB_KEY = re.compile(
    r""" ^
    [+]
    (?P<keylen> 0 | [1-9][0-9]* )
    ,
    (?P<vallen> 0 | [1-9][0-9]* )
    :
    (?P<rest> .* )
    $ """,
    re.X | re.S,
)


def cdb_line(match: re.Match[str]) -> str:
    """Parse a single line of `cdb -d` output."""
    keylen, vallen = int(match.group("keylen")), int(match.group("vallen"))
    if keylen < 1 or vallen < 1:
        sys.exit(f"cdbdump: klen {keylen} or vlen {vallen} cannot be zero")
    dump = match.group("rest")
    if len(dump) < keylen + 2 + vallen + 1:
        sys.exit(f"cdbdump: klen {keylen} vlen {vallen} rlen {len(dump)}")

    key, sep, val, eol, dump = (
        dump[:keylen],
        dump[keylen : keylen + 2],
        dump[keylen + 2 : keylen + 2 + vallen],
        dump[keylen + 2 + vallen : keylen + 2 + vallen + 1],
        dump[keylen + 2 + vallen + 1 :],
    )
    if not (key.endswith(NULL) and sep == "->" and eol == "\n"):
        sys.exit(f"cdbdump: bad: {key!r} {sep!r} {val!r} {eol!r} {dump!r}")

    key = key[:-1]
    name_enc = key
    name = []
    while name_enc:
        glen = name_enc[0].encode(defs.SAFEENC)[0]
        if len(name_enc) < glen + 1:
            sys.exit(f"cdbdump: key {key!r} glen {glen}")
        name.append(name_enc[1 : glen + 1])
        name_enc = name_enc[glen + 1 :]

    rectype = val[:2].encode(defs.SAFEENC)
    print(f"key {key!r} name {'.'.join(name)} record type {rectype[0] * 256 + rectype[1]}")

    return dump


def test_tinydns_data(cfg: defs.Config, tempd: pathlib.Path) -> None:
    """Test the creation of the data.cdb file."""
    print("\n==== test_tinydns_data")
    data = tempd / "data"
    data_cdb = data.with_suffix(".cdb")
    if not data.is_file():
        sys.exit(f"Expected {data} to be a file")
    if data_cdb.exists():
        sys.exit(f"Did not expect {data_cdb} to exist")

    tdata = cfg.sbindir / "tinydns-data"
    outp = cfg.check_output([tdata])
    if outp:
        sys.exit(f"{tdata!r} produced output: {outp!r}")

    if not data_cdb.is_file():
        sys.exit(f"{tdata!r} did not create {data_cdb}")

    dump = cfg.check_output(["cdb", "-d", "--", data_cdb], encoding=defs.SAFEENC)
    while match := RE_CDB_KEY.match(dump):
        dump = cdb_line(match)

    if dump != "\n":
        sys.exit(f"cdbdump: leftover {dump!r}")



"""Test the `tinydns` and `dnsq` programs."""

from __future__ import annotations

import dataclasses
import enum
import os
import socket
import struct
import subprocess  # noqa: S404
import sys
import time
import typing

from . import defs
from . import test_get


if typing.TYPE_CHECKING:
    import pathlib
    from collections.abc import Callable


QUERY_PROGRESS = 10000

PORT_DNS = 53


class ProcessFDType(str, enum.Enum):
    """File descriptor types, enough for this very limited test."""

    IPV4 = "IPv4"


@dataclasses.dataclass(frozen=True)
class ProcessFDInfo:
    """Information about a single file descriptor in a process."""

    fileno: int
    access: str
    lock: str
    ftype: ProcessFDType


@dataclasses.dataclass(frozen=True)
class ProcessSocketFDInfo(ProcessFDInfo):
    """Information about a single socket in a process."""

    proto: str
    address: str
    port: int


@dataclasses.dataclass(frozen=True)
class ProcessInfo:
    """Information about a process and (some of) its file descriptors."""

    pid: int
    pgid: int
    tid: int
    comm: str

    uid: int
    username: str

    files: list[ProcessFDInfo]


@dataclasses.dataclass
class ParseProcessInfo:
    """Parse the output of `lsof -F` for our very limited testing needs."""

    cfg: defs.Config
    cproc: dict[str, str] = dataclasses.field(default_factory=dict)
    cfile: dict[str, str] = dataclasses.field(default_factory=dict)
    cfiles: list[ProcessFDInfo] = dataclasses.field(default_factory=list)
    result: dict[int, ProcessInfo] = dataclasses.field(default_factory=dict)

    def finish_file_ipv4(self) -> ProcessFDInfo:
        """Store the information about a socket."""
        fields = self.cfile["addrport"].split(":", 1)
        return ProcessSocketFDInfo(
            fileno=int(self.cfile["fileno"]),
            access=self.cfile["access"],
            lock=self.cfile["lock"],
            ftype=ProcessFDType.IPV4,
            proto=self.cfile["proto"],
            address=fields[0],
            port=int(fields[1]),
        )

    def finish_file(self) -> None:
        """Store the information about the current file descriptor."""
        if self.cfile["type"] == "IPv4":
            fdata = self.finish_file_ipv4()
        else:
            sys.exit(f"Internal error: unhandled file type: {self.cfile!r}")

        self.cfile.clear()

        self.cfiles.append(fdata)

    def finish_process(self) -> None:
        """Store the information about the current process."""
        if self.cfile:
            self.finish_file()

        pdata = ProcessInfo(
            pid=int(self.cproc["pid"]),
            pgid=int(self.cproc["pgid"]),
            tid=int(self.cproc["tid"]),
            comm=self.cproc["comm"],
            uid=int(self.cproc["uid"]),
            username=self.cproc["username"],
            files=list(self.cfiles),
        )

        self.cfiles.clear()
        self.cproc.clear()

        if pdata.pid in self.result:
            sys.exit(
                f"lsof output a duplicate process set for {pdata.pid}: "
                f"first {self.result[pdata.pid]!r} now {pdata!r}",
            )
        self.result[pdata.pid] = pdata

    def parse(self, lines: list[str]) -> dict[int, ProcessInfo]:
        """Parse some lsof output lines."""

        def h_process(_first: str, rest: str) -> None:
            """Start parsing a new process set."""
            if self.cproc:
                self.finish_process()

            self.cproc["pid"] = rest

        def h_file(_first: str, rest: str) -> None:
            """Start parsing a new file descriptor set."""
            if self.cfile:
                self.finish_file()

            self.cfile["fileno"] = rest

        handlers: dict[str, Callable[[str, str], None] | tuple[dict[str, str], str]] = {
            "p": h_process,
            "g": (self.cproc, "pgid"),
            "R": (self.cproc, "tid"),
            "c": (self.cproc, "comm"),
            "u": (self.cproc, "uid"),
            "L": (self.cproc, "username"),
            "f": h_file,
            "a": (self.cfile, "access"),
            "l": (self.cfile, "lock"),
            "t": (self.cfile, "type"),
            "P": (self.cfile, "proto"),
            "n": (self.cfile, "addrport"),
        }

        for line in lines:
            if not line:
                sys.exit("lsof output an empty line")
            first, rest = line[0], line[1:]
            handler = handlers.get(first)
            if not handler:
                continue

            if isinstance(handler, tuple):
                handler[0][handler[1]] = rest
            else:
                handler(first, rest)

        if self.cproc:
            self.finish_process()

        return self.result


def check_single_udp_socket(processes: dict[int, ProcessInfo], pid: int, ipaddr: str) -> bool:
    """Make sure tinydns is only listening on a single UDP socket."""
    pdata = processes.get(pid)
    if pdata is None or len(pdata.files) != 1:
        return False

    data = pdata.files[0]
    if not isinstance(data, ProcessSocketFDInfo):
        return False

    return data.proto == "UDP" and data.address == ipaddr and data.port == PORT_DNS


def run_tinydns(
    cfg: defs.Config,
    tempd: pathlib.Path,
    callback: Callable[[subprocess.Popen[str], str], None],
) -> None:
    """Test the `tinydns` and `dnsq` programs."""
    ipaddr = defs.RECORDS[0].address
    svcdir = tempd / defs.SVCDIR
    if not svcdir.is_dir():
        sys.exit(f"Expected {svcdir} to be a directory")
    os.chdir(svcdir)

    with subprocess.Popen(  # noqa: S603
        ["./run"],
        shell=False,
        env=cfg.subenv,
        bufsize=0,
        encoding=defs.MINENC,
        stdout=subprocess.PIPE,
    ) as proc:
        try:
            print(f"- spawned process {proc.pid}")
            assert proc.stdout is not None  # noqa: S101  # mypy needs this
            print("- waiting for the 'starting tinydns' line")
            line = proc.stdout.readline()
            print(f"- got line {line!r}")
            if line != "starting tinydns\n":
                sys.exit(f"Unexpected first line from tinydns: {line!r}")

            lines = cfg.check_output(
                [
                    "lsof",
                    "-a",
                    "-n",
                    "-P",
                    "-p",
                    str(proc.pid),
                    "-F",
                    "-i4udp:53",
                ],
            ).splitlines()
            cfg.diag(f"lsof output: {lines!r}")
            processes = ParseProcessInfo(cfg).parse(lines)
            if not check_single_udp_socket(processes, proc.pid, ipaddr):
                sys.exit(f"tinydns should listen on a single UDP socket: {processes!r}")
            print(f"- tinydns is listening at {ipaddr}:53")
            callback(proc, ipaddr)
        finally:
            print("- we spawned a process, checking if it has exited")
            res = proc.poll()
            if res is None:
                print(f"Terminating process {proc.pid}")
                proc.terminate()
                time.sleep(0.5)
                res = proc.poll()
                if res is None:
                    print(f"Killing process {proc.pid}")
                    proc.kill()

    res = proc.wait()
    print(f"Process {proc.pid}: exit code {res}")


def test_tinydns_run(cfg: defs.Config, tempd: pathlib.Path) -> None:
    """Test the `tinydns` and `dnsq` programs."""
    print("\n==== test_tinydns_run")
    if os.geteuid() != 0:
        print("- not running as root, skipped")
        return

    dnsq = cfg.bindir / "dnsq"
    if not dnsq.is_file() or not os.access(dnsq, os.X_OK):
        sys.exit(f"Not an executable file: {dnsq}")

    def test_dnsq(proc: subprocess.Popen[str], ipaddr: str) -> None:
        """Run `dnsq` against the specified `tinydns` instance."""
        assert proc.stdout is not None  # noqa: S101  # mypy needs this

        for rec in defs.RECORDS:
            rdef = defs.TYPES[rec.rtype]
            print(f"- querying {ipaddr} for {rdef.query} {rec.name}")
            test_get.test_tinydns_query(
                cfg,
                [
                    "timelimit",
                    "-p",
                    "-t10",
                    "-T10",
                    "--",
                    dnsq,
                    rdef.query,
                    rec.name,
                    ipaddr,
                ],
                rec,
            )

            print("  - there should be a single line of output from tinydns")
            line = proc.stdout.readline()
            cfg.diag(f"line: {line!r}")

    run_tinydns(cfg, tempd, test_dnsq)


def test_tinydns_run_udp(cfg: defs.Config, tempd: pathlib.Path) -> None:
    """Test the `tinydns` program by sending UDP packets."""
    print("\n==== test_tinydns_run_udp")
    if os.geteuid() != 0:
        print("- not running as root, skipped")
        return

    def test_dnsq(proc: subprocess.Popen[str], ipaddr: str) -> None:
        """Run `dnsq` against the specified `tinydns` instance."""
        assert proc.stdout is not None  # noqa: S101  # mypy needs this

        pkt = struct.pack(
            ">HHHHHH8p4p1pHH",
            17,
            0,
            1,
            0,
            0,
            0,
            "example".encode("us-ascii"),
            "com".encode("us-ascii"),
            b"",
            1,
            1,
        )
        print(f"- query packet: {pkt!r}")

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.connect((ipaddr, 53))

        for idx in range(QUERY_PROGRESS * 100):
            pkt = struct.pack(">H", (idx + 17) % 65536) + pkt[2:]
            sent = sock.send(pkt)
            if sent != len(pkt):
                sys.exit(f"Only sent {sent} of {len(pkt)} bytes")

            line = proc.stdout.readline()
            if "+" not in line:
                sys.exit(f"Query {idx + 1}: expected a '+' in the tinydns log line {line!r}")

            resp = sock.recv(4096)
            if resp[:2] != pkt[:2]:
                sys.exit(f"Query {idx + 1}: unexpected ID: query {pkt!r} response {resp!r}")
            if resp[4:8] != b"\x00\x01\x00\x01":
                sys.exit(
                    f"Query {idx + 1}: unexpected query/response count: "
                    f"query {pkt!r} response {resp!r}",
                )

            if (idx % QUERY_PROGRESS) == QUERY_PROGRESS - 1:
                print(f"- sent {idx + 1} queries to {ipaddr}")

            time.sleep(0.00001)

    run_tinydns(cfg, tempd, test_dnsq)



"""Test the `tinydns-edit` tool and the generated data file."""

from __future__ import annotations

import os
import sys
import typing

from . import defs


if typing.TYPE_CHECKING:
    import pathlib


# This function may be simplified in varous ways with later versions of Python
def test_tinydns_edit(cfg: defs.Config, tempd: pathlib.Path) -> None:  # noqa: C901
    """Test the operation of `tinydns-edit`."""
    print("\n==== test_tinydns_edit")

    data = tempd / "data"
    data_temp = data.with_suffix(".tmp")
    for chk in (data, data_temp):
        if chk.exists():
            sys.exit(f"Did not expect {chk} to exist")

    tedit = cfg.sbindir / "tinydns-edit"
    if not tedit.is_file() or not os.access(tedit, os.X_OK):
        sys.exit(f"Not an executable file: {tedit}")

    data.write_text("", encoding=defs.MINENC)
    old_lines: list[str] = []

    for idx, rec in enumerate(defs.RECORDS):
        cfg.diag(f"- {idx}: {rec}")
        cmd: list[pathlib.Path | str] = [
            tedit,
            data,
            data_temp,
            "add",
            rec.rtype,
            rec.name,
            rec.address,
        ]
        cfg.diag("  - " + " ".join(map(str, cmd)))

        outp = cfg.check_output(cmd)
        if outp:
            sys.exit(f"{cmd!r} produced output: {outp!r}")
        if data_temp.exists():
            sys.exit(f"{cmd!r} left {data_temp}")
        if not data.exists():
            sys.exit(f"{cmd!r} removed {data}")

        lines = data.read_text(encoding=defs.MINENC).splitlines()
        if len(lines) != idx + 1:
            sys.exit(f"{cmd!r} left {len(lines)} instead of {idx + 1} lines: {lines!r}")
        if lines[:-1] != old_lines:
            sys.exit(f"{cmd!r} did not preserve {old_lines!r}: {lines!r}")

        line = lines[-1]
        fields = line.split(":")
        if len(fields) < 3:  # noqa: PLR2004  # this may be better with py310 `match`/`case`
            sys.exit(f"{cmd!r} added a short line: {line!r}")
        exp = defs.TYPES[rec.rtype].prefix + rec.name
        if fields[0] != exp:
            sys.exit(f"{cmd!r} added an unexpected line: {line!r}")
        if rec.address not in fields[1:]:
            sys.exit(f"{cmd!r} added a line with no {rec.address!r}: {line!r}")

        old_lines.append(line)

    print("After the tinydns-edit test the contents of the data file is:")
    print(data.read_text(encoding=defs.MINENC), end="")



"""Test the `tinydns-conf` tool and the generated directory."""

from __future__ import annotations

import os
import pwd
import sys
import typing

from . import defs


if typing.TYPE_CHECKING:
    import pathlib


TINYDNS = "/usr/sbin/tinydns"


def test_tinydns_conf(cfg: defs.Config, tempd: pathlib.Path) -> None:
    """Test the operation of `tinydns-conf`."""
    print("\n==== test_tinydns_conf")
    ipaddr = defs.RECORDS[0].address
    svcacct = pwd.getpwuid(os.getuid()).pw_name
    svcdir = tempd / defs.SVCDIR
    if svcdir.exists():
        sys.exit(f"Did not expect {svcdir} to exist")

    tconf = cfg.sbindir / "tinydns-conf"
    if not tconf.is_file() or not os.access(tconf, os.X_OK):
        sys.exit(f"Not an executable file: {tconf}")

    cmd: list[pathlib.Path | str] = [
        tconf,
        svcacct,
        svcacct,
        svcdir,
        ipaddr,
    ]
    outp = cfg.check_output(cmd)
    if outp:
        sys.exit(f"{cmd!r} produced output: {outp!r}")

    cfg.check_call(["find", "--", svcdir, "-ls"])

    if not svcdir.is_dir():
        sys.exit(f"{cmd!r} did not create {svcdir}")

    ipfile = svcdir / "env/IP"
    if not ipfile.is_file():
        sys.exit(f"{cmd!r} did not create {ipfile}")
    iptext = ipfile.read_text(encoding=defs.MINENC)
    cfg.diag(f"{ipfile}: {iptext!r}")
    if iptext != ipaddr + "\n":
        sys.exit(f"{cmd!r} created {ipfile} with weird contents: {iptext!r}")

    runfile = svcdir / "run"
    if not runfile.is_file():
        sys.exit(f"{cmd!r} did not create {runfile}")
    lines = runfile.read_text(encoding=defs.MINENC).splitlines()
    cfg.diag(f"{runfile} last line: {lines[-1]!r}")
    if lines[0] != "#!/bin/sh" or not lines[-1].endswith(" " + TINYDNS):
        sys.exit(f"{cmd!r} created unexpected {runfile}: {lines!r}")

    lines[-1] = lines[-1].replace(" " + TINYDNS, " " + str(cfg.sbindir / "tinydns"))
    runfile.write_text("".join(line + "\n" for line in lines), encoding=defs.MINENC)
    cfg.check_call(["cat", "--", runfile])

    data_cdb = svcdir / "root/data.cdb"
    data_cdb.write_bytes((tempd / "data.cdb").read_bytes())
    cfg.check_call(["stat", "--", data_cdb])



"""Test the operation of the tinydns utilities."""

from __future__ import annotations

import argparse
import contextlib
import dataclasses
import errno
import os
import pathlib
import sys
import tempfile

from . import defs
from . import test_conf
from . import test_data
from . import test_edit
from . import test_get
from . import test_run


@dataclasses.dataclass(frozen=True)
class Config(defs.Config):
    """Runtime configuration for the test runner."""

    skip_run_test: bool


def parse_arguments() -> Config:
    """Parse the command-line arguments."""
    parser = argparse.ArgumentParser(prog="tinytest")
    parser.add_argument(
        "-b",
        "--bindir",
        type=pathlib.Path,
        required=True,
        help="the path to the djbdns user tools",
    )
    parser.add_argument(
        "-s",
        "--sbindir",
        type=pathlib.Path,
        required=True,
        help="the path to the djbdns system tools",
    )
    parser.add_argument(
        "-S",
        "--skip-run-test",
        action="store_true",
        help="skip the test that runs tinydns",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="verbose operation; display diagnostic output",
    )

    args = parser.parse_args()

    subenv = dict(os.environ)
    subenv["LC_ALL"] = "C.UTF-8"
    with contextlib.suppress(KeyError):
        del subenv["LANGUAGE"]

    return Config(
        bindir=args.bindir.resolve(),
        sbindir=args.sbindir.resolve(),
        skip_run_test=args.skip_run_test,
        subenv=subenv,
        verbose=args.verbose,
    )


def run_tests(cfg: Config, tempd: pathlib.Path) -> None:
    """Run some tests."""
    cfg.diag(f"Running tests in {tempd}")
    os.chdir(tempd)

    test_edit.test_tinydns_edit(cfg, tempd)
    test_data.test_tinydns_data(cfg, tempd)
    test_get.test_tinydns_get(cfg, tempd)
    test_conf.test_tinydns_conf(cfg, tempd)

    if not cfg.skip_run_test:
        test_run.test_tinydns_run(cfg, tempd)
        test_run.test_tinydns_run_udp(cfg, tempd)
    else:
        print("\n==== Skipping test_tinydns_run*")

    print("\n==== All fine!")


def fixup_tempd_group(cfg: Config, tempd: pathlib.Path) -> None:
    """Change a temporary directory's group if needed."""
    testdir = tempd / "test-perms"
    testdir.mkdir()
    cfg.diag(f"Trying to set access mode 3755 on {testdir}")
    failed = False
    try:
        testdir.chmod(0o3755)
    except OSError as err:
        failed = True
        cfg.diag(f"- failed: {err}")
        if err.errno not in {errno.EPERM, errno.EACCES}:
            raise

    if not failed:
        cfg.diag("Everything seems to be in order")
        return

    cfg.diag(f"Removing {testdir}")
    testdir.rmdir()

    cfg.diag("Examining the situation")
    tempd_stat = tempd.stat()
    tempd_gid = tempd_stat.st_gid
    our_gid = os.getgid()
    cfg.diag(f"- tempd gid: {tempd_gid} our gid: {our_gid}")
    if tempd_gid == our_gid:
        sys.exit("Don't know how to handle things with the same group ID")
    cfg.diag(f"Trying to change the group ID of {tempd} from {tempd_gid} to {our_gid}")
    os.chown(tempd, tempd_stat.st_uid, our_gid)

    cfg.diag("Now retrying the test")
    testdir.mkdir()
    testdir.chmod(0o3755)
    testdir.rmdir()
    cfg.diag(f"Looks like we fixed the access permissions of {tempd}")


def main() -> None:
    """Parse options, run the tests."""
    cfg = parse_arguments()
    with tempfile.TemporaryDirectory() as tempd_obj:
        tempd = pathlib.Path(tempd_obj).absolute()
        cfg.diag(f"Using temporary directory {tempd}")
        fixup_tempd_group(cfg, tempd)
        try:
            run_tests(cfg, tempd)
        finally:
            try:
                os.chdir("/")
            except OSError as err:
                print(f"Could not chdir to /: {err}", file=sys.stderr)


if __name__ == "__main__":
    main()



"""Test the `tinydns-get` tool."""

from __future__ import annotations

import os
import re
import sys
import typing

from . import defs


if typing.TYPE_CHECKING:
    import pathlib


RE_FIRST = re.compile(
    r""" ^
    (?P<rtype> 0 | [1-9][0-9]* )
    \s+
    (?P<name> [A-Za-z0-9/.]+ )
    :
    $ """,
    re.X,
)

RE_SECOND = re.compile(
    r""" ^
    (?: 0 | [1-9][0-9]* ) \s+ bytes
    , \s+
        (?: 0 | [1-9][0-9]* ) \+
        (?: 0 | [1-9][0-9]* ) \+
        (?: 0 | [1-9][0-9]* ) \+
        (?: 0 | [1-9][0-9]* ) \s+ records
    (?:
        , \s+
        [a-z]+
    )*
    $ """,
    re.X,
)

RE_THIRD = re.compile(
    r""" ^
    query: \s+
    (?P<rtype> 0 | [1-9][0-9]* )
    \s+
    (?P<name> [A-Za-z0-9/.]+ )
    $ """,
    re.X,
)


def test_tinydns_query(
    cfg: defs.Config,
    cmd: list[pathlib.Path | str],
    rec: defs.TestRecord,
) -> None:
    """Run a test query and check its output."""
    rdef = defs.TYPES[rec.rtype]
    print(
        f"  - expect {len(rec.answers)} "
        f"answer{'' if len(rec.answers) == 1 else 's'} for "
        f"{rdef.query} {rec.name}",
    )

    cfg.diag("- " + " ".join(map(str, cmd)))
    lines = cfg.check_output(cmd).splitlines()
    if len(lines) < 3 + len(rec.answers):
        sys.exit(f"{cmd!r}: too few lines: {lines!r}")

    match = RE_FIRST.match(lines[0])
    if not (match and int(match.group("rtype")) == rdef.rtype and match.group("name") == rec.name):
        sys.exit(f"{cmd!r}: bad first line: {lines!r}")

    match = RE_SECOND.match(lines[1])
    if not match:
        sys.exit(f"{cmd!r}: bad second line: {lines!r}")

    match = RE_THIRD.match(lines[2])
    if not (match and int(match.group("rtype")) == rdef.rtype and match.group("name") == rec.name):
        sys.exit(f"{cmd!r}: bad query line: {lines!r}")

    ans_lines = lines[3 : 3 + len(rec.answers)]
    more_lines = lines[3 + len(rec.answers) :]
    cfg.diag(f"  - answer lines: {ans_lines!r}")
    cfg.diag(f"  - additional lines: {more_lines!r}")

    if ans_lines != rec.answers:
        sys.exit(
            f"{cmd!r}: bad answers: expected {rec.answers!r}, got {ans_lines!r}, full {lines!r}",
        )

    if rec.answers[0].startswith("authority: "):
        okay = all(line.startswith("additional: ") for line in more_lines)
    else:
        okay = all(line.startswith(("additional: ", "authority: ")) for line in more_lines)
    if not okay:
        sys.exit(f"{cmd!r}: superfluous response: {lines!r}")


def test_tinydns_get(cfg: defs.Config, _tempd: pathlib.Path) -> None:  # noqa: PT019
    """Test the operation of `tinydns-get`."""
    print("\n==== test_tinydns_get")

    tget = cfg.bindir / "tinydns-get"
    if not tget.is_file() or not os.access(tget, os.X_OK):
        sys.exit(f"Not an executable file: {tget}")

    for rec in defs.RECORDS:
        rdef = defs.TYPES[rec.rtype]
        test_tinydns_query(cfg, [tget, rdef.query, rec.name], rec)
